#include <f8/internal/pch.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>

#include <mutex>
#include <condition_variable>
#include <atomic>
#include <thread>

#include <a8/uuid.h>

#include <f8/udplog.h>
#include <f8/app.h>
#include <f8/msgqueue.h>
#include <f8/protoutils.h>
#include <f8/tglog.h>
#include <f8/httpclientpool.h>
#include <a8/queue.h>
#include <f8/timer.h>
#include <f8/userapp.h>
#include <f8/iomgr.h>
#include <f8/threadpool.h>
#include <f8/msghdr.h>

static const int MAX_ZONE_ID = 100;
static const int MAX_NODE_ID = 8;
static const int MAX_INSTANCE_ID = 500;

static const int MAX_SYS_HTTP_NUM = 2;
static const int MAX_USER_HTTP_NUM = 8;
static const int MAX_ALL_HTTP_NUM = MAX_SYS_HTTP_NUM + MAX_USER_HTTP_NUM;

static const char* const PROJ_ROOT_FMT = "/data/logs/%s";
static const char* const PROJ_LOG_ROOT_FMT = "/data/logs/%s/logs";
static const char* const PROJ_LOG_FILENAME_FMT = "log_$pid_%Y%m%d.log";

namespace f8
{

    class App::Impl
    {
    private:
        UserApp* user_app_ = nullptr;
        int argc_ = 0;
        char** argv_ = nullptr;
        int exit_code_ = 0;
        volatile bool terminated_ = false;
        int nowtime_ = 0;
        long long max_run_delay_time_ = 0;

        int zone_id_ = 0;
        int node_id_ = 0;
        int instance_id_ = 0;
        std::set<int> flags_;

        f8::ThreadPool preprocess_thread_pool_;
        a8::Queue net_msg_queue_;
        std::atomic<long long> msgnode_size_ = {0};
        std::atomic<long long> working_msgnode_size_ = {0};

        std::shared_ptr<a8::uuid::SnowFlake> uuid_;
        std::mutex *loop_mutex_ = nullptr;
        std::condition_variable *loop_cond_ = nullptr;

    public:

        bool Init()
        {
            signal(SIGPIPE, SIG_IGN);
            srand(time(nullptr));
            nowtime_ = time(nullptr);

            if (!ParseOpt()) {
                exit_code_ = 1;
                return false;
            }

            uuid_ = std::make_shared<a8::uuid::SnowFlake>();
            loop_mutex_ = new std::mutex();
            loop_cond_ = new std::condition_variable();

            uuid_->SetMachineId((node_id_ - 1) * MAX_NODE_ID + instance_id_);
            InitLog();
            f8::IoMgr::Instance()->Init();
            f8::MsgQueue::Instance()->Init();
            f8::Timer::Instance()->Init();
            f8::TGLog::Instance()->Init(user_app_->GetPkgName(), false, 0);
            f8::HttpClientPool::Instance()->Init(MAX_ALL_HTTP_NUM, MAX_SYS_HTTP_NUM, MAX_USER_HTTP_NUM);
            user_app_->Init();
            preprocess_thread_pool_.Init
                (std::max(1, user_app_->GetPreprocessThreadNum()),
                 [this] (f8::ThreadPool::Context* ctx)
                 {
                     while (!ctx->Terminated()) {
                         list_head* work_list = ctx->GetWorkList();
                         while (!list_empty(work_list)){
                             MsgHdr* hdr = list_first_entry(work_list, MsgHdr, entry);
                             list_del_init(&hdr->entry);
                             user_app_->PreprocessSocketMsg(hdr);
                             net_msg_queue_.Push(&hdr->entry);
                             NotifyLoopCond();
                         }
                         ctx->Sleep(1000);
                     }
                 });
            return true;
        }

        void UnInit()
        {
            user_app_->UnInit();
            f8::TGLog::Instance()->UnInit();
            f8::HttpClientPool::Instance()->UnInit();
            f8::Timer::Instance()->UnInit();
            f8::MsgQueue::Instance()->UnInit();
            f8::IoMgr::Instance()->UnInit();
            UnInitLog();

            delete loop_cond_;
            loop_cond_ = nullptr;
            delete loop_mutex_;
            loop_mutex_ = nullptr;
        }

        int Run(int argc, char* argv[], UserApp* user_app)
        {
            argc_ = argc;
            argv_ = argv;
            user_app_ = user_app;
            int delta_time = 0;
            if (Init()) {
                a8::tick_t last_stat_tick = a8::XGetTickCount();
                while (!Terminated()) {
                    a8::tick_t begin_tick = a8::XGetTickCount();
                    nowtime_ = time(nullptr);
                    f8::Timer::Instance()->Update();
                    f8::MsgQueue::Instance()->Update();
                    DispatchNetMsg();
                    user_app->Update(delta_time);
                    f8::IoMgr::Instance()->Update();
                    a8::tick_t end_tick = a8::XGetTickCount();
                    if (end_tick - begin_tick > max_run_delay_time_) {
                        max_run_delay_time_ = end_tick - begin_tick;
                    }
                    Schedule();
                    end_tick = a8::XGetTickCount();
                    if (end_tick - last_stat_tick > 0) {
                        delta_time = end_tick - last_stat_tick;
                        last_stat_tick = end_tick;
                    } else {
                        delta_time = 0;
                    }
                }
                UnInit();
            }
            return exit_code_;
        }

        bool HasFlag(int flag)
        {
            return flags_.find(flag) != flags_.end();
        }

        const std::string GetPkgName()
        {
            return user_app_->GetPkgName();
        }

        void NotifyLoopCond()
        {
            std::unique_lock<std::mutex> lk(*loop_mutex_);
            loop_cond_->notify_all();
        }

        long long NewNodeUuid()
        {
            return uuid_->Generate();
        }

        const std::string NewGlobalUuid()
        {
            std::string id = a8::Format("%d%d", {100 + zone_id_, uuid_->Generate()});
            return id;
        }

        void Schedule()
        {
            std::unique_lock<std::mutex> lk(*loop_mutex_);
            bool has_task = false;
            {
                net_msg_queue_.Fetch();
                if (!list_empty(net_msg_queue_.GetWorkList())) {
                    has_task = true;
                }
            }
            if (!has_task) {
                has_task = user_app_->HasTask();
            }
            if (!has_task) {
                int sleep_time = f8::Timer::Instance()->GetIdleTime();
                loop_cond_->wait_for(lk, std::chrono::milliseconds(sleep_time));
            } else {
                int sleep_time = 1;
                loop_cond_->wait_for(lk, std::chrono::milliseconds(sleep_time));
            }
        }

        void DispatchNetMsg()
        {
            net_msg_queue_.Fetch();
            list_head* work_list = net_msg_queue_.GetWorkList();
            while (!list_empty(work_list)){
                MsgHdr* hdr = list_first_entry(work_list, MsgHdr, entry);
                list_del_init(&hdr->entry);
                user_app_->DispatchSocketMsg(hdr);
                --msgnode_size_;
                MsgHdr::Destroy(hdr);
            }
        }

        void AddSocketMsg(unsigned short sockfrom,
                          long long socket_handle,
                          int msgid,
                          int seqid,
                          const char *msgbody,
                          int bodylen,
                          const char *playload,
                          int playload_len,
                          unsigned short tag
                          )
        {
            char *p = (char*)malloc(sizeof(MsgHdr) + bodylen);
            MsgHdr* hdr = (MsgHdr*)p;
            hdr->sockfrom = sockfrom;
            hdr->seqid = seqid;
            hdr->msgid = msgid;
            hdr->socket_handle = socket_handle;
            hdr->buflen = bodylen;
            hdr->tag = tag;
            if (bodylen > 0) {
                memmove(p + sizeof(MsgHdr), msgbody, bodylen);
            }
            ++msgnode_size_;
            preprocess_thread_pool_.PostMsg(socket_handle, &hdr->entry);
        }

        bool ParseOpt()
        {
            int ch = 0;
            while ((ch = getopt(argc_, argv_, "z:n:i:f:")) != -1) {
                switch (ch) {
                case 'z':
                    {
                        zone_id_ = a8::XValue(optarg);
                    }
                    break;
                case 'n':
                    {
                        node_id_ = a8::XValue(optarg);
                    }
                    break;
                case 'i':
                    {
                        instance_id_ = a8::XValue(optarg);
                    }
                    break;
                case 'f':
                    {
                        std::vector<std::string> strings;
                        a8::Split(optarg, strings, ',');
                        for (auto& str : strings) {
                            flags_.insert(a8::XValue(str).GetInt());
                        }
                    }
                    break;
                }
            }
            if (zone_id_ <= 0) {
                a8::XPrintf("启动失败,缺少-z参数\n", {});
                return false;
            } else if (node_id_ > MAX_ZONE_ID) {
                a8::XPrintf("启动失败,-z参数不能大于%d\n", {MAX_ZONE_ID});
                return false;
            } else if (node_id_ <= 0) {
                a8::XPrintf("启动失败,缺少-n参数\n", {});
                return false;
            } else if (node_id_ > MAX_NODE_ID) {
                a8::XPrintf("启动失败,-n参数不能大于%d\n", {MAX_NODE_ID});
                return false;
            } else if (instance_id_ <= 0) {
                a8::XPrintf("启动失败,缺少-i参数\n", {});
                return false;
            } else if (instance_id_ > MAX_INSTANCE_ID) {
                a8::XPrintf("启动失败,-i参数不能大于%d\n", {MAX_INSTANCE_ID});
                return  false;
            }
            a8::XPrintf("starting zone_id:%d node_id:%d instance_id:%d pid:%d\n",
                        {
                         zone_id_,
                         node_id_,
                         instance_id_,
                         GetPid()
                        });
            return true;
        }

        int GetPid()
        {
            return getpid();
        }

        int GetZoneId()
        {
            return zone_id_;
        }

        int GetNodeId()
        {
            return node_id_;
        }

        int GetInstanceId()
        {
            return instance_id_;
        }

        void InitLog()
        {
            std::string filename_fmt = PROJ_LOG_FILENAME_FMT;
            a8::ReplaceString(filename_fmt, "$pid",
                              a8::XValue(f8::App::Instance()->GetPid()));

            std::string proj_root_dir = a8::Format
                (PROJ_ROOT_FMT, {user_app_->GetPkgName()});
            std::string proj_log_root_dir = a8::Format
                (PROJ_LOG_ROOT_FMT, {user_app_->GetPkgName()});
            std::string log_file_name = a8::Format
                (PROJ_LOG_ROOT_FMT, {user_app_->GetPkgName()}) + "/" + filename_fmt;

            a8::MkDir(proj_root_dir);
            a8::MkDir(proj_log_root_dir);
            a8::XPrintf("log_file_name:%s\n", {log_file_name});
            f8::UdpLog::Instance()->SetLogFileName(log_file_name);
            f8::UdpLog::Instance()->Init();
            f8::UdpLog::Instance()->Info("proj_root_dir:%s", {proj_root_dir});
            f8::UdpLog::Instance()->Info("proj_log_root_dir:%s", {proj_log_root_dir});
            f8::UdpLog::Instance()->Info("log_file_name:%s", {log_file_name});
        }

        void UnInitLog()
        {
            f8::UdpLog::Instance()->UnInit();
        }

        bool Terminated()
        {
            return terminated_;
        };

    };

    App::App():impl_(std::make_unique<Impl>())
    {

    }

    int App::Run(int argc, char* argv[], UserApp* user_app)
    {
        impl_ = std::make_unique<Impl>();
        return impl_->Run(argc, argv, user_app);
    }

    bool App::HasFlag(int flag)
    {
        return impl_->HasFlag(flag);
    }

    const std::string App::GetPkgName()
    {
        return impl_->GetPkgName();
    }

    void App::NotifyLoopCond()
    {
        impl_->NotifyLoopCond();
    }

    long long App::NewNodeUuid()
    {
        return impl_->NewNodeUuid();
    }

    const std::string App::NewGlobalUuid()
    {
        return impl_->NewGlobalUuid();
    }

    int App::GetPid()
    {
        return impl_->GetPid();
    }

    int App::GetZoneId()
    {
        return impl_->GetZoneId();
    }

    int App::GetNodeId()
    {
        return impl_->GetNodeId();
    }

    int App::GetInstanceId()
    {
        return impl_->GetInstanceId();
    }

    void App::AddSocketMsg(unsigned short sockfrom,
                           long long socket_handle,
                           int msgid,
                           int seqid,
                           const char *msgbody,
                           int bodylen,
                           const char *playload,
                           int playload_len,
                           unsigned short tag)
    {
        impl_->AddSocketMsg(sockfrom,
                            socket_handle,
                            msgid,
                            seqid,
                            msgbody,
                            bodylen,
                            playload,
                            playload_len,
                            tag
                            );
    }

}
